/***************************************************************************
 *   Copyright (C) 2007 by David Lopez Vinacua                             *
 *   david@bajavista.com                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

/* * * * * * * * * * * * * * * * * * * * * *
 *  dcview.c                               *
 *  Code and Data for dnscache log viewer  *
 * * * * * * * * * * * * * * * * * * * * * *
 *
 * Created:
 *  260907 David Lopez
 * Modified:
 *
 */

#include <stdio.h>
#include <time.h>
#include <string.h>

#define LogPath "/service/dnscache/log/main"
#define Current "current"

typedef unsigned char Byte;

short daltool_GetHexByte(char *src)
/* Gets a two-digit hex byte from src. Returns -1 if fails */
{
  Byte          v,w;

  if (!src || !(*src))
     return -1;
  if ( (*src>='0') && (*src<='9') )
     v=(*src)-'0';
  else if ( (*src>='a') && (*src<='f') )
     v=(*src)-'a'+10;
  else if ( (*src>='A') && (*src<='F') )
     v=(*src)-'A'+10;
  else
     return -1;
  src++;
  if (!*src)
     return v;
  if ( (*src>='0') && (*src<='9') )
     w=(*src)-'0';
  else if ( (*src>='a') && (*src<='f') )
     w=(*src)-'a'+10;
  else if ( (*src>='A') && (*src<='F') )
     w=(*src)-'A'+10;
  else
     return -1;
  return 16*v+w;
}

char *daltool_GetTs(unsigned int T)
/* Returns 32bit sec time representation in text */
{
  static char       kt_s[16];
  
  kt_s[0]='\0';
  if (T<60)
     sprintf(kt_s,"%us",T);
  else if (T<3600)
     {
     if (T%60)
        sprintf(kt_s,"%um%02us",T/60,(T%60));
     else
        sprintf(kt_s,"%um",T/60);
     }
  else if (T<3600*24)
     {
     if (T%60)
        sprintf(kt_s,"%uh%02um%02us",T/3600,(T/60)%60,T%60);
     else if ((T/60)%60)
        sprintf(kt_s,"%uh%02um",T/3600,(T/60)%60);
     else
        sprintf(kt_s,"%uh",T/3600);
     }
  else
     {
     if (T%60)
        sprintf(kt_s,"%ud%uh%02um%02us",T/86400,(T/3600)%24,(T/60)%60,T%60);
     else if ((T/60)%60)
        sprintf(kt_s,"%ud%uh%02um",T/86400,(T/3600)%24,(T/60)%60);
     else if ((T/3600)%24)
        sprintf(kt_s,"%ud%uh",T/86400,(T/3600)%24);
     else
        sprintf(kt_s,"%ud",T/86400);
     }
  return (char *)&kt_s[0];
}


short GetHexu32(char *src, unsigned int *res)
/* Gets an unsigned 32bit Hex BigEndian value, returns 0 on success */
{
  unsigned int    v;
  short           r,i;
  
  if (!src)
     return -1;
  v=0;
  i=0;
  while (i<4)
    {
    r=daltool_GetHexByte(src);
    if (r>=0)
       {
       v=256*v+r;
       src+=2;
       }
    else
       return -2;
    i++;
    }
  *res=v;
  return 0;
}

short GetHexu16(char *src, unsigned short *res)
/* Gets an unsigned 16bit Hex BigEndian value, returns 0 on success */
{
  unsigned short  v;
  short           r,i;
  
  if (!src)
     return -1;
  v=0;
  i=0;
  while (i<2)
    {
    r=daltool_GetHexByte(src);
    if (r>=0)
       {
       v=256*v+r;
       src+=2;
       }
    else
       return -2;
    i++;
    }
  *res=v;
  return 0;
}

int GetNum(char *s)
/* Parses an unsigned value from s */
{
  unsigned int    v;
  
  if (!s)
     return -1;
  v=0;
  while ( (*s>='0') && (*s<='9') )
    {
    v=10*v+((*s)-'0');
    s++;
    }
  return v;
}

char *Parse_Type(unsigned short Type)
/* Returns query type */
{
  static char     s[8];
  
  switch (Type)
    {
    case 0x0001: return "A";
    case 0x0002: return "NS";
    case 0x0005: return "CNAME";
    case 0x0006: return "SOA";
    case 0x000c: return "PTR";
    case 0x000f: return "MX";
    case 0x0010: return "TXT";
    case 0x001b: return "GPOS";
    case 0x001c: return "AAAA";
    case 0x0026: return "A6";
    case 0x0063: return "SPF";
    case 0x00fb: return "IXFR";
    case 0x00fc: return "AXFR";
    case 0x00ff: return "*";
    default: 
      sprintf(s,"[0x%04x]",Type);
      return &s[0];
    }
}

short Parse(FILE *f, FILE *o)
/* Parses given file to standard output */
{
  char            s[256];
  unsigned short  cp,pp;
  unsigned int    i1,i2;
  unsigned short  w1,w2;
  int             i;
  time_t          t;
  struct tm      *tm;
  
  if (f)
     {
     while ( (fgets(&s[0],sizeof(s),f)) )
       {
       cp=0;
       // Line starts with the time stamp
       if (s[0]=='@')
          {
          cp++;
          pp=cp;
          while (s[cp] && (s[cp]!=' '))
            cp++;
          // Timestamp is 24 hex digits => 3x32bit values
          // It seems to use 32bit nanosecond value at the end (9 digits)
          // The rest may be a 64bit time_t, but only lower 32bit seem valid
          // So we ignore first 32bit value, by now
          if ( (cp-pp==24) && s[cp] )
             {
             if (GetHexu32(&s[pp+8],&i1) || GetHexu32(&s[pp+16],&i2))
                fprintf(o,"**Invalid line (wrong timestamp)**\n");
             else
                {
                t=i1;
                tm=localtime(&t);
                fprintf(o,"%04u%02u%02u-%02u:%02u:%02u.%06u ",
                 1900+tm->tm_year,1+tm->tm_mon,tm->tm_mday,tm->tm_hour,tm->tm_min,tm->tm_sec,i2/1000);
                // Now comes the logged message
                cp++;
                pp=cp;
                cp=strlen(&s[0]);
                if (cp && (s[cp-1]=='\n'))
                   s[cp-1]='\0';
                switch (s[pp])
                  {
                  case 'c': // cached
                    pp+=7;
                    // May have a number (type of record needed), or (cname | ns | nxdomain)
                    switch (s[pp])
                      {
                      case 'c': // cname
                        fprintf(o,"cached %s\n",&s[pp]);
                        break;
                      case 'n': // ns | nxdomain
                        switch (s[pp+1])
                          {
                          case 's': // ns
                          case 'x': // nxdomain
                            fprintf(o,"cached %s\n",&s[pp]);
                            break;
                          }
                        break;
                      default:  // Must be a type number
                        i=GetNum(&s[pp]);
                        fprintf(o,"cached %s",Parse_Type(i));
                        while (s[pp] && (s[pp]!=' '))
                          pp++;
                        if (s[pp])
                           fprintf(o,"%s\n",&s[pp]);
                      }
                    break;
                  case 'd': // drop
                    fprintf(o,"%s\n",&s[pp]);
                    break;
                  case 'l': // lame
                    pp+=5;
                    fprintf(o,"lame %u.%u.%u.%u %s\n",
                     daltool_GetHexByte(&s[pp]),daltool_GetHexByte(&s[pp+2]),daltool_GetHexByte(&s[pp+4]),daltool_GetHexByte(&s[pp+6]),&s[pp+9]);
                    break;
                  case 'n': // nodata | nxdomain
                    switch (s[pp+1])
                      {
                      case 'o': // nodata
                        pp+=7;
                        i=GetNum(&s[pp+9]);
                        fprintf(o,"nodata %u.%u.%u.%u TTL=%s",
                          daltool_GetHexByte(&s[pp]),daltool_GetHexByte(&s[pp+2]),daltool_GetHexByte(&s[pp+4]),daltool_GetHexByte(&s[pp+6]),
                          daltool_GetTs(i));
                        pp+=10;
                        while (s[pp] && (s[pp]!=' '))
                          pp++;
                        if (s[pp])
                           {
                           pp++;
                           i=GetNum(&s[pp]);
                           pp++;
                           while (s[pp] && (s[pp]!=' '))
                             pp++;
                           fprintf(o," %s%s\n",Parse_Type(i),&s[pp]);
                           }
                        break;
                      case 'x': // nxdomain
                        pp+=9;
                        i=GetNum(&s[pp+9]);
                        fprintf(o,"nxdomain %u.%u.%u.%u TTL=%s",
                          daltool_GetHexByte(&s[pp]),daltool_GetHexByte(&s[pp+2]),daltool_GetHexByte(&s[pp+4]),daltool_GetHexByte(&s[pp+6]),
                          daltool_GetTs(i));
                        pp+=10;
                        while (s[pp] && (s[pp]!=' '))
                          pp++;
                        if (s[pp])
                           fprintf(o,"%s\n",&s[pp]);
                        break;
                      }
                    break;
                  case 'q': // query
                    pp+=6;
                    i=GetNum(&s[pp]);
                    fprintf(o,"query sn:%u",i);
                    pp++;
                    while (s[pp] && (s[pp]!=' '))
                      pp++;
                    if (s[pp] && !GetHexu16(&s[pp+10],&w1) && !GetHexu16(&s[pp+15],&w2))
                       {
                       pp++;
                       fprintf(o," %u.%u.%u.%u:%u id=%u",
                         daltool_GetHexByte(&s[pp]),daltool_GetHexByte(&s[pp+2]),daltool_GetHexByte(&s[pp+4]),daltool_GetHexByte(&s[pp+6]),
                         w1,w2);
                       pp+=19;
                       i=GetNum(&s[pp]);
                       pp++;
                       while (s[pp] && (s[pp]!=' '))
                         pp++;
                       fprintf(o," %s%s\n",Parse_Type(i),&s[pp]);
                       }
                    break;
                  case 'r': // rr
                    pp+=3;
                    i=GetNum(&s[pp+9]);
                    fprintf(o,"rr %u.%u.%u.%u TTL=%s",
                      daltool_GetHexByte(&s[pp]),daltool_GetHexByte(&s[pp+2]),daltool_GetHexByte(&s[pp+4]),daltool_GetHexByte(&s[pp+6]),
                      daltool_GetTs(i));
                    pp+=9;
                    while (s[pp] && (s[pp]!=' '))
                      pp++;
                    pp++;
                    switch (s[pp])
                      {
                      case 'n': // ns
                        fprintf(o," %s\n",&s[pp]);
                        break;
                      case 's': // soa
                        cp=pp;
                        pp+=4;
                        while (s[pp] && (s[pp]!=' '))
                          pp++;
                        pp++;
                        while (s[pp] && (s[pp]!=' '))
                          pp++;
                        s[pp]='\0';
                        pp++;
                        fprintf(o," %s ",&s[cp]);
                        cp=pp;
                        while (s[pp] && (s[pp]!=' ') && (s[pp]!='.') )
                          pp++;
                        if (s[pp]=='.')
                           {
                           s[pp]='\0';
                           fprintf(o,"%s@",&s[cp]);
                           pp++;
                           cp=pp;
                           while (s[pp] && (s[pp]!=' ') )
                             pp++;
                           s[pp-1]='\0';
                           fprintf(o,"%s ",&s[cp]);
                           pp++;
                           i=GetNum(&s[pp]);
                           i1=i;
                           while (s[pp] && (s[pp]!=' ') )
                             pp++;
                           pp++;
                           i=GetNum(&s[pp]);
                           while (s[pp] && (s[pp]!=' ') )
                             pp++;
                           pp++;
                           fprintf(o,"Serial: %u Refresh: %s ",i1,daltool_GetTs(i));
                           i=GetNum(&s[pp]);
                           fprintf(o,"Retry: %s ",daltool_GetTs(i));
                           while (s[pp] && (s[pp]!=' ') )
                             pp++;
                           pp++;
                           i=GetNum(&s[pp]);
                           fprintf(o,"Expires: %s ",daltool_GetTs(i));
                           while (s[pp] && (s[pp]!=' ') )
                             pp++;
                           pp++;
                           i=GetNum(&s[pp]);
                           fprintf(o,"Min: %s\n",daltool_GetTs(i));
                           }
                        break;
                      default:
                        if ( (s[pp]>='0') && (s[pp]<='9') )
                           {  // This seems to be <type> <name> <IP>
                           i=GetNum(&s[pp]);
                           pp++;
                           while (s[pp] && (s[pp]!=' '))
                             pp++;
                           if (s[pp])
                              {
                              pp++;
                              cp=pp+1;
                              while (s[cp] && (s[cp]!=' '))
                                cp++;
                              s[cp]='\0';
                              cp++;
                              if (i==28)  // IPv6
                                 {
                                 fprintf(o," %s %s ",Parse_Type(i),&s[pp]);
                                 fprintf(o,"%c%c",s[cp],s[cp+1]);
                                 for (i=1; i<8; i++)
                                   fprintf(o,":%c%c%c%c",s[cp+4*i],s[cp+4*i+1],s[cp+4*i+2],s[cp+4*i+3]);
                                 fprintf(o,"\n");
                                 }
                              else
                                 fprintf(o," %s %s %u.%u.%u.%u\n",Parse_Type(i),&s[pp],
                                   daltool_GetHexByte(&s[cp]),daltool_GetHexByte(&s[cp+2]),daltool_GetHexByte(&s[cp+4]),daltool_GetHexByte(&s[cp+6]));
                              }
                           }
                        else
                           fprintf(o,"**NYI** %s\n",&s[pp]);
                      }
                    break;
                  case 's': // sent | servfail | starting | stats
                    switch (s[pp+3])
                      {
                      case 't': // sent | stats
                        if (s[pp+1]=='e') // sent
                           {
                           pp+=5;
                           i=GetNum(&s[pp]);
                           i1=i;
                           while (s[pp] && (s[pp]!=' '))
                             pp++;
                           fprintf(o,"sent sn:%u%sB\n",i1,&s[pp]);
                           }
                        else
                           {  // stats
                           pp+=6;
                           i=GetNum(&s[pp]);
                           while (s[pp] && (s[pp]!=' '))
                             pp++;
                           i1=i;
                           if (s[pp])
                              {
                              pp++;
                              i=GetNum(&s[pp]);
                              i2=i;
                              while (s[pp] && (s[pp]!=' '))
                                pp++;
                              pp++;
                              i=GetNum(&s[pp]);
                              w1=i;
                              while (s[pp] && (s[pp]!=' '))
                                pp++;
                              pp++;
                              i=GetNum(&s[pp]);
                              fprintf(o,"stats qCnt=%u ChdB=%u pUDP=%u pTCP=%u\n",i1,i2,w1,i);
                              }
                           }
                        break;
                      case 'v': // servfail
                        fprintf(o,"%s\n",&s[pp]);
                        break;
                      case 'r': // starting
                        fprintf(o,"%s\n",&s[pp]);
                        break;
                      }
                    break;
                  case 't': // tcpopen | tcpclose | tx
                    if (s[pp+1]=='x') // tx
                       {
                       pp+=3;
                       i=GetNum(&s[pp]);
                       i1=i;
                       while (s[pp] && (s[pp]!=' '))
                         pp++;
                       pp++;
                       i=GetNum(&s[pp]);
                       w1=i;
                       while (s[pp] && (s[pp]!=' '))
                         pp++;
                       pp++;
                       cp=pp;
                       while (s[cp] && (s[cp]!=' '))
                         cp++;
                       s[cp]='\0';
                       fprintf(o,"tx gl=%u %s %s",i1,Parse_Type(w1),&s[pp]);
                       pp=cp+1;
                       cp=pp;
                       while (s[cp] && (s[cp]!=' '))
                         cp++;
                       s[cp]='\0';
                       fprintf(o," AuthDom:%s AuthSrvs:",&s[pp]);
                       s[cp]=' ';
                       while (s[cp])
                         {
                         cp++;
                         fprintf(o," %u.%u.%u.%u",
                           daltool_GetHexByte(&s[cp]),daltool_GetHexByte(&s[cp+2]),daltool_GetHexByte(&s[cp+4]),daltool_GetHexByte(&s[cp+6]));
                         cp+=8;
                         }
                       fprintf(o,"\n");
                       }
                    else
                       {  // tcpopen | tcpclose
                       
                       }
                    break;
                  }
                }
             }
          else
             fprintf(o,"**Invalid line (bad timestamp)**\n");
          }
       else
          fprintf(o,"**Invalid line**\n");
       }
     fclose(f);
     }
  return 0;
}

int main(int argc, char *argv[])
{
  char            fn[192];
  FILE           *f,*o;
  
  if ( (argc>1) && (strcmp(argv[1],"-")==0) )
     f=stdin;
  else
     {
     sprintf(fn,"%s/%s",LogPath,Current);
     f=fopen(&fn[0],"r");
     }
  o=stdout;
  Parse(f,o);
  return 0;
}
